# sandboxr/backends/docker.py

import subprocess
import tempfile
import shutil
import os
import uuid
from typing import List

class DockerSandbox:
    def __init__(self, packages: List[str], base_dir=None):
        self.packages = packages or []
        self.id = uuid.uuid4().hex[:8]
        self.image_tag = f"sandboxr_{self.id}"
        self.build_dir = base_dir or tempfile.mkdtemp(prefix="sandboxr_docker_")
        self.dockerfile = os.path.join(self.build_dir, "Dockerfile")
        # Create output directory
        self.output_dir = os.path.join(self.build_dir, "output")
        os.makedirs(self.output_dir, exist_ok=True)

    def create(self):
        df = ["FROM python:3.9-slim", "WORKDIR /sandbox"]
        if self.packages:
            df.append("RUN pip install " + " ".join(self.packages))
        with open(self.dockerfile, "w") as f:
            f.write("\n".join(df))
        subprocess.check_call(
            ["docker", "build", "--quiet", "-t", self.image_tag, self.build_dir],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )
        self.ready = True
        return self

    def is_ready(self):
        return getattr(self, "ready", False)

    def exec(self, code: str, timeout=30):
        cmd = ["docker", "run", "--rm", self.image_tag, "python", "-c", code]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        try:
            out, err = proc.communicate(timeout=timeout)
            return out.decode(), err.decode(), proc.returncode
        except subprocess.TimeoutExpired:
            proc.kill()
            return "", "Timed out", -1

    def exec_file(self, filepath: str, timeout=30):
        abs_path = os.path.abspath(filepath)
        cmd = [
            "docker", "run", "--rm",
            "-v", f"{abs_path}:/sandbox/script.py",
            "-v", f"{self.output_dir}:/sandbox/output",
            self.image_tag, "python", "/sandbox/script.py"
        ]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        try:
            out, err = proc.communicate(timeout=timeout)
            return out.decode(), err.decode(), proc.returncode
        except subprocess.TimeoutExpired:
            proc.kill()
            return "", "Timed out", -1

    def install(self, packages: List[str]):
        """
        Append a RUN layer to the Dockerfile for the new packages,
        then remove the old image and rebuild quietly.
        """
        with open(self.dockerfile, "a") as f:
            f.write("\nRUN pip install " + " ".join(packages))

        subprocess.check_call(
            ["docker", "rmi", "-f", self.image_tag],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )
        subprocess.check_call(
            ["docker", "build", "--quiet", "-t", self.image_tag, self.build_dir],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )

    def teardown(self):
        subprocess.call(["docker", "rmi", "-f", self.image_tag],
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL)
        shutil.rmtree(self.build_dir, ignore_errors=True)
        self.ready = False



# pysand/manager.py
from .backends.virtualenv import VirtualenvSandbox
from .backends.docker import DockerSandbox

class SandboxManager:
    def __init__(self, backend: str, packages=None, base_dir=None):
        self.backend = backend
        self.packages = packages or []
        self.base_dir = base_dir

    def create(self):
        if self.backend == "virtualenv":
            return VirtualenvSandbox(self.packages, base_dir=self.base_dir).create()
        elif self.backend == "docker":
            return DockerSandbox(self.packages, base_dir=self.base_dir).create()
        else:
            raise ValueError(f"Unknown backend: {self.backend}")



# sandboxr/cli.py

import argparse
import sys
from .manager import SandboxManager

def main():
    p = argparse.ArgumentParser(prog="sandboxr")
    p.add_argument("--backend", choices=["virtualenv","docker"], required=True)
    p.add_argument("--packages", default="", help="comma-separated list")
    p.add_argument("--timeout", "-t", type=int, default=30, help="Max seconds to allow code to run (default: 30)")
    grp = p.add_mutually_exclusive_group(required=True)
    grp.add_argument("--code", help="Python code string")
    grp.add_argument("--file", help="Path to .py file")
    p.add_argument("--keep", action="store_true", help="Do not teardown")
    args = p.parse_args()

    pkgs = args.packages.split(",") if args.packages else []
    mgr = SandboxManager(args.backend, pkgs)
    sandbox = mgr.create()

    if args.code:
        out, err, code = sandbox.exec(args.code, timeout=args.timeout)
    else:
        out, err, code = sandbox.exec_file(args.file, timeout=args.timeout)

    if out:
        # Print to stdout
        sys.stdout.write(out)
    if err:
        # Print errors to stderr
        sys.stderr.write(err)

    if not args.keep:
        sandbox.teardown()

    sys.exit(code)

if __name__ == "__main__":
    main()



# tests/test_sandbox.py
import pytest
from sandboxr.manager import SandboxManager

@pytest.fixture(params=["virtualenv", "docker"])
def mgr(request):
    return SandboxManager(backend=request.param, packages=["pytest"])

def test_create_and_teardown(mgr, tmp_path):
    sandbox = mgr.create()
    assert sandbox.is_ready()
    sandbox.teardown()
    assert not sandbox.is_ready()

def test_run_simple_code(mgr):
    sandbox = mgr.create()
    out, err, code = sandbox.exec("print('hello sandboxr')")
    assert code == 0
    assert out.strip() == "hello sandboxr"
    sandbox.teardown()

def test_install_additional_package(mgr):
    sandbox = mgr.create()
    sandbox.install = getattr(sandbox, "install", None)
    # virtualenv has install; docker installs at create time
    if sandbox.install:
        sandbox.install(["requests"])
    out, err, code = sandbox.exec("import requests; print(True)")
    assert code == 0
    sandbox.teardown()

def test_timeout(mgr):
    sandbox = mgr.create()
    out, err, code = sandbox.exec("import time; time.sleep(5)", timeout=1)
    assert code != 0
    sandbox.teardown()

def test_run_file(tmp_path, mgr):
    sandbox = mgr.create()
    script = tmp_path / "script.py"
    script.write_text("print(2 + 3)")
    out, err, code = sandbox.exec_file(str(script))
    assert code == 0
    assert out.strip() == "5"
    sandbox.teardown()



# add a new file tests/test_cli.py:
import subprocess, sys, pytest
@pytest.mark.parametrize("args,expected", [
    (["--backend","virtualenv","--code","print(1+1)"], "2"),
    (["--backend","virtualenv","--packages","requests","--code","import requests;print(True)"], "True"),
])
def test_sandboxr_cli(tmp_path, args, expected):
    cmd = [sys.executable, "-m", "sandboxr.cli", *args]
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.stdout.strip() == expected
    assert result.returncode == 0



# sandboxr/backends/virtualenv.py

import venv
import subprocess
import shutil
import tempfile
import os
from typing import List

class VirtualenvSandbox:
    def __init__(self, packages: List[str], base_dir=None):
        self.packages = packages or []
        self.base_dir = base_dir or tempfile.mkdtemp(prefix="sandboxr_")
        self.env_dir = os.path.join(self.base_dir, "env")

    def create(self):
        venv.create(self.env_dir, with_pip=True)
        self.python = os.path.join(self.env_dir, "bin", "python")
        if self.packages:
            subprocess.check_call(
                [self.python, "-m", "pip", "install", *self.packages],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )
        self.ready = True
        return self

    def is_ready(self):
        return getattr(self, "ready", False)

    def exec(self, code: str, timeout=30):
        proc = subprocess.Popen(
            [self.python, "-c", code],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        try:
            out, err = proc.communicate(timeout=timeout)
            return out.decode(), err.decode(), proc.returncode
        except subprocess.TimeoutExpired:
            proc.kill()
            return "", "Timed out", -1

    def exec_file(self, filepath: str, timeout=30):
        proc = subprocess.Popen(
            [self.python, filepath],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        try:
            out, err = proc.communicate(timeout=timeout)
            return out.decode(), err.decode(), proc.returncode
        except subprocess.TimeoutExpired:
            proc.kill()
            return "", "Timed out", -1

    def install(self, packages: List[str]):
        """
        Install additional packages into an existing virtualenv sandbox.
        """
        subprocess.check_call(
            [self.python, "-m", "pip", "install", *packages],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )

    def teardown(self):
        shutil.rmtree(self.base_dir, ignore_errors=True)
        self.ready = False



from setuptools import setup, find_packages

# Read in the README for the long description on PyPI
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="sandboxr",
    version="0.1.3",
    author="Ani Kulkarni",
    author_email="aniruddha.k1911@gmail.com",
    description="Quickly spin up isolated Python sandboxes (virtualenv or Docker), install dependencies, execute code, and tear down.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/akulka404/sandboxr",
    packages=find_packages(),
    install_requires=[],
    entry_points={
        "console_scripts": [
            "sandboxr=sandboxr.cli:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
    python_requires=">=3.6",
    keywords=["sandbox", "virtualenv", "docker", "automation", "cli"],
    include_package_data=True,
    zip_safe=False,
)
